/* Process model C form file: ip_host_arp.pr.c */
/* Portions of this file Copyright (C) MIL 3, Inc. 1992 */



/* OPNET system definitions */
#include <opnet.h>
#include "ip_host_arp.pr.h"
FSM_EXT_DECS



/* Header block */
#include "ip_arp_sup.h"
#include <sys/file.h>

typedef struct
{
	int			node_num;
	int			phys_addr;
} arp_entry;
                 

/* Transition Mnemonics */
#define ARRIVAL		op_intrpt_type () == OPC_INTRPT_STRM

/* identification constant for IP protocol within Ethernet frames */
#define NET_PROT_IP		0x800

#define MAC_STREAM      0



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	arp_entry*	             		sv_addr_table;
	int	                    		sv_num_entries;
	char	                   		sv_addr_file [512];
	Objid	                  		sv_my_id;
	Ici*	                   		sv_mac_iciptr;
	arp_table*	             		sv_at_ptr;
	} ip_host_arp_state;

#define pr_state_ptr            		((ip_host_arp_state*) SimI_Mod_State_Ptr)
#define addr_table              		pr_state_ptr->sv_addr_table
#define num_entries             		pr_state_ptr->sv_num_entries
#define addr_file               		pr_state_ptr->sv_addr_file
#define my_id                   		pr_state_ptr->sv_my_id
#define mac_iciptr              		pr_state_ptr->sv_mac_iciptr
#define at_ptr                  		pr_state_ptr->sv_at_ptr




/* Process model interrupt handling procedure */



void
ip_host_arp ()
	{
	int				instrm, next_node;
	char			str0 [512], str1 [512];
	Packet*			pkptr;
	Ici*			iciptr;
	int				phys_addr;
	int				status;
	char			error_string [512];


	FSM_ENTER (ip_host_arp)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (INIT) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "INIT")
			{
			/******************************************************************************************************
			** Process Model : ip_host_arp                                                                       **
			** State         : INIT                                                                              **
			** Authors       : Ravi Malghan (rmalghan@descartes.gmu.edu)                                         **
			**                                                                                                   **
			** This state loads the memory table that is specified in the configuration file into the memory and **
			** creates an ICI pointer to communicate with layer 2 (FDDI MAC Layer).                              **
			**                                                                                                   **
			******************************************************************************************************/
			
			my_id = op_id_self ();                /*       Obtain the object ID of the ARP processor.          */
			
			/* 
			** Obtain the name of the address conversion file to be used by this ARP processor. Load this address
			** file into the memory
			*/
			op_ima_obj_attr_get (my_id, "addr_file", addr_file);
			at_ptr = arp_sup_table_setup (addr_file);
			 
			mac_iciptr = op_ici_create ("ip_mac_req");           /* Create an Ici to communicate with layer 2. */
			}


		/** state (INIT) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "INIT")
			{
			}


		/** state (INIT) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (WAIT) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "WAIT")
			{
			
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,ip_host_arp)


		/** state (WAIT) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "WAIT")
			{
			}


		/** state (WAIT) transition processing **/
		FSM_TRANSIT_ONLY ((ARRIVAL), 2, state2_enter_exec, ;, "WAIT")
		/*---------------------------------------------------------*/



		/** state (ARRIVAL) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "ARRIVAL")
			{
			/************************************************************************************************************
			** Process Model : ip_host_arp                                                                             **
			** State         : ARRIVAL                                                                                 **
			**                                                                                                         **
			** Transition to this state occurs when there is a packet arrival in the wait state. If no ICI is received **
			** that corresponds to this packet simulation is aborted. If there is an ICI, next node number is obtained **
			** from  the ICI and corresponding  physical address is  determined using the  address table loaded in the **
			** memory. If  the address translation is done, then  the destination address is  installed in the ICI and **
			** the ICI packet is sent to mac layer. This state on default transits to WAIT state.                      **
			**                                                                                                         **
			************************************************************************************************************/
			
			instrm = op_intrpt_strm();               /* Determine on which input stream the packet is arrived. */
			
			/* 
			** Obtain the packet and the accompanying ICI. If there is no such ICI end the simulation.
			*/
			pkptr = op_pk_get( instrm );
			iciptr = op_intrpt_ici();  
			if( iciptr == OPC_NIL )
			{
				sprintf( error_string, "Simulation aborted; error in object (%d)", op_id_self () );
				op_sim_end( error_string, "ip_arp: required ICI not received", " ", " " );
			}
			
			op_ici_attr_get( iciptr, "next_node", &next_node );           /* Obtain the next node number. */
			
			/* 
			** Determine the corresponding physical address. 
			*/
			status = arp_sup_phys_addr_get( at_ptr, next_node, &phys_addr );
			
			if( status == OPC_COMPCODE_FAILURE )
			{
				/* If the address translation failed, issue an error message. */
				sprintf( str0, "ARP: discarding unmappable packet (%d)", op_pk_id( pkptr ) );
				sprintf( str1, "node number: (%d)", next_node );
				op_pk_destroy( pkptr );
				op_sim_message( str0, str1 );	
			}
			else
			{	
				/* place the destination address and protocol type (IP) into the ICI */
				op_ici_attr_set( mac_iciptr, "dest_addr", phys_addr );
				op_ici_attr_set( mac_iciptr, "protocol_type", NET_PROT_IP );
			
				/* send the packet coupled with the ICI */
				op_ici_install( mac_iciptr );
				op_pk_send( pkptr, MAC_STREAM );
			}
			}


		/** state (ARRIVAL) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "ARRIVAL")
			{
			}


		/** state (ARRIVAL) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,ip_host_arp)
	}




void
ip_host_arp_svar (prs_ptr,var_name,var_p_ptr)
	ip_host_arp_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (ip_host_arp_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("addr_table" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_addr_table);
	if (Vos_String_Equal ("num_entries" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_num_entries);
	if (Vos_String_Equal ("addr_file" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_addr_file);
	if (Vos_String_Equal ("my_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_id);
	if (Vos_String_Equal ("mac_iciptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_mac_iciptr);
	if (Vos_String_Equal ("at_ptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_at_ptr);

	FOUT;
	}




void
ip_host_arp_diag ()
	{
	int				instrm, next_node;
	char			str0 [512], str1 [512];
	Packet*			pkptr;
	Ici*			iciptr;
	int				phys_addr;
	int				status;
	char			error_string [512];

	FIN (ip_host_arp_diag ())


	FOUT;
	}




void
ip_host_arp_terminate ()
	{
	int				instrm, next_node;
	char			str0 [512], str1 [512];
	Packet*			pkptr;
	Ici*			iciptr;
	int				phys_addr;
	int				status;
	char			error_string [512];

	FIN (ip_host_arp_terminate ())


	FOUT;
	}



Compcode
ip_host_arp_init (pr_state_pptr)
	ip_host_arp_state			**pr_state_pptr;
	{
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	FIN (ip_host_arp_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (ip_host_arp)",
			sizeof (ip_host_arp_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (ip_host_arp_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



